/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* Copyright (C) 2011 - DIGITEO - Michael Baudin                             */
/* Copyright (C) 2011 - National Institute of Informatics - Benoit Goepfert  */
/*                                                                           */
/* This file must be used under the terms of the CeCILL.                     */
/* This source file is licensed as described in the file COPYING, which      */
/* you should have received as part of this distribution.  The terms         */
/* are also available at                                                     */
/* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt                  */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

// From ANSI C
#include "stdio.h"
#include "float.h"
#include "stdlib.h"


/*API scilab includes*/
#include "stack-c.h"
#include "api_scilab.h" 
#include "Scierror.h"
#include "localization.h"
#include "sciprint.h"

/*Gateway includes*/
#include "gw_spilu_support.h"

// From this lib
#include "conv.h"

// Private functions
int spilu_Double2IntegerArgument ( char * fname , int ivar , double dvalue , int * ivalue );


// Get a scalar double.
//
// Arguments
// fname (input): the name of the function
// ivar (input): the index of the input argument
// rhs (input): the number of input arguments
// defaultdouble (input): the default value of the double.
// mydouble (output): the required double
//
// Description
// If there is a problem, produces an error and return 0.
// If there is no problem, returns 1.
//
// If rhs < ivar, then set the default value.
// If rhs >= ivar, and the value is [], then set the default value.
// If not, then set the actual value.
//
int spilu_getScalarDouble (char *fname, int ivar, int rhs, double defaultdouble, double * mydouble)
{
	int iRows, iCols;
	int dblStkPtr=0;
	double * value;

	if (Rhs>=ivar) {
		if (VarType(ivar)==1)
		{
			GetRhsVar(ivar,"d",&iRows,&iCols,&dblStkPtr);
			if (iRows==0) 
			{
				*mydouble = defaultdouble;
			}
			else
			{
				if ((iRows!=1)||(iCols!=1))
				{
					Scierror(501,"%s: Wrong size for input argument #%d: %d-by-%d matrix expected.\n",fname,ivar,1,1);
					return(0);
				}
				value = stk(dblStkPtr);
				*mydouble = *value;
			}
		} else {
			Scierror(501,"%s: Wrong type for argument %d: Matrix expected.\n",fname,ivar);
			return 0;
		}
	} else {
		*mydouble = defaultdouble;
	}
	return 1;
}

// Check that a double is in a given range
//
// Arguments
// fname (input): the name of the function
// ivar (input): the index of the input argument
// mydouble (input): the double to check
// mindouble (input): the minimum value of the double.
// maxdouble (input): the maximum value of the double.
//
// Description
// If myint is not in range, produce an error and returns 1.
// If myint is OK, returns 0.
//
int spilu_checkDoubleInRange (char *fname, int ivar, double mydouble, double mindouble, double maxdouble)
{
	if ( mydouble > maxdouble )
	{
		Scierror(204,_("%s: Wrong value for input argument #%d: Must be < %f.\n"),fname, ivar, maxdouble);
		return 0;
	}
	if ( mydouble < mindouble )
	{
		Scierror(204,_("%s: Wrong value for input argument #%d: Must be > %f.\n"),fname, ivar, mindouble);
		return 0;
	}
	return 1;
}

// Get a scalar integer from a Double.
//
// Arguments
// fname (input): the name of the function
// ivar (input): the index of the input argument
// rhs (input): the number of input arguments
// defaultint (input): the default value of the int.
// myint (output): the required int
//
// Description
// If there is a problem, produces an error and return 0.
// If there is no problem, returns 1.
//
// If rhs < ivar, then set the default value.
// If rhs >= ivar, and the value is [], then set the default value.
// If not, then set the actual value.
//
int spilu_getScalarIntegerFromScalarDouble (char *fname, int ivar, int rhs, int defaultint, int * myint)
{
	int iRows, iCols;
	int dblStkPtr=0;
	double * value;
	int readFlag;

	if (rhs>=ivar) {
		if (VarType(ivar)==1)
		{
			GetRhsVar(ivar,"d",&iRows,&iCols,&dblStkPtr);
			if (iRows==0) 
			{
				*myint = defaultint;
			}
			else
			{
				if ((iRows!=1)||(iCols!=1))
				{
					Scierror(501,"%s: Wrong size for input argument #%d: %d-by-%d matrix expected.\n",fname,ivar,1,1);
					return(0);
				}
				value = stk(dblStkPtr);
				readFlag = spilu_Double2IntegerArgument ( fname , ivar , *value , myint );
				if ( !readFlag )
				{
					return 0;
				}
			}
		} else {
			Scierror(501,"%s: Wrong type for argument %d: Matrix expected.\n",fname,ivar);
			return 0;
		}
	} else {
		*myint = defaultint;
	}
	return 1;
}

// Check that an integer is in a given range
//
// Arguments
// fname (input): the name of the function
// ivar (input): the index of the input argument
// myint (input): the integer to check
// minint (input): the minimum value of the integer.
// maxint (input): the maximum value of the integer.
//
// Description
// If myint is not in range, produce an error and returns 1.
// If myint is OK, returns 0.
//
int spilu_checkIntegerInRange (char *fname, int ivar, int myint, int minint, int maxint)
{
	if ( myint > maxint )
	{
		Scierror(204,_("%s: Wrong value for input argument #%d: Must be < %d.\n"),fname, ivar, maxint);
		return 0;
	}
	if ( myint < minint )
	{
		Scierror(204,_("%s: Wrong value for input argument #%d: Must be > %d.\n"),fname, ivar, minint);
		return 0;
	}
	return 1;
}

// 
// splspc_Double2IntegerArgument --
//   Convert the given double into an integer.
//
// Arguments
//   fname (input) : the name of the Scilab function generating this error
//   ivar (input) : the index of the input variable
//   dvalue (input) : the source double value
//   ivalue (output) : the target integer value
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
// Check that the double is in the integer range.
// Check that the double has no fractional part.
//
int spilu_Double2IntegerArgument ( char * fname , int ivar , double dvalue , int * ivalue )
{
	if ( dvalue > INT_MAX ) {
		Scierror(999,_("%s: Too large integer value in argument #%d: found %e while maximum value is %d.\n"),fname,ivar , dvalue , INT_MAX );
		return 0;
	}
	if ( dvalue < INT_MIN ) {
		Scierror(999,_("%s: Too large integer value in argument #%d: found %e while minimum value is %d.\n"),fname,ivar , dvalue , INT_MIN );
		return 0;
	}
	*ivalue = (int)dvalue;
	// Now check that the double was really an integer, with 
	// a zero fractionnal part.
	if ( (double)*ivalue != dvalue ) {
		Scierror(999,_("%s: Wrong integer value in argument #%d: found %e which is different from the closest integer %d.\n"),fname,ivar , dvalue , *ivalue );
		return 0;
	}

	return 1;
}

// 
// spilu_getArgumentDrop --
//   Get the drop parameter.
//
// Arguments
//   fname (input) : the name of the Scilab function generating this error
//   ivar (input) : the index of the input variable
//   rhs (input): the number of input arguments
//   A (input) : the sparse matrix
//   drop (output) : the value
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
// Check that drop is positive.
// If drop is not provided or is equal to [], then 
// set the default value, drop=0.001*max(abs(A)).
//
int spilu_getArgumentDrop ( char * fname , int ivar , int rhs, SciSparse A , double * drop )
{
	double defaultdrop;
	int readFlag;

	defaultdrop = 0.001*spiluc_eltm(A);
	readFlag = spilu_getScalarDouble (fname, ivar, Rhs, defaultdrop, drop);
	if (readFlag==0)
	{
		return 0;
	}
	readFlag = spilu_checkDoubleInRange (fname, ivar, *drop, 0, DBL_MAX);
	if (readFlag==0)
	{
		return 0;
	}
	return 1;
}

// 
// spilu_getArgumentAlpha --
//   Get the alpha parameter.
//
// Arguments
//   fname (input) : the name of the Scilab function generating this error
//   ivar (input) : the index of the input variable
//   rhs (input): the number of input arguments
//   alpha (output) : the alpha value
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
// Check that drop is positive.
// If drop is not provided or is equal to [], then 
// set the default value, alpha=0.5.
//
int spilu_getArgumentAlpha ( char * fname , int ivar , int rhs, double * alpha )
{
	double defaultalpha;
	int readFlag;

	defaultalpha = 0.5;

	readFlag = spilu_getScalarDouble (fname, ivar, rhs, defaultalpha, alpha);
	if (readFlag==0)
	{
		return 0;
	}
	readFlag = spilu_checkDoubleInRange (fname, ivar, *alpha, 0, DBL_MAX);
	if (readFlag==0)
	{
		return 0;
	}
	return 1;
}

// 
// spilu_getArgumentPermtol --
//   Get the permtol parameter.
//
// Arguments
//   fname (input) : the name of the Scilab function generating this error
//   ivar (input) : the index of the input variable
//   rhs (input): the number of input arguments
//   permtol (output) : the value
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
// Check that drop is positive.
// If drop is not provided or is equal to [], then 
// set the default value, permtol=0.5.
//
int spilu_getArgumentPermtol ( char * fname , int ivar , int rhs, double * permtol )
{
	double defaultpermtol;
	int readFlag;

	defaultpermtol = 0.5;

	readFlag = spilu_getScalarDouble (fname, ivar, rhs, defaultpermtol, permtol);
	if (readFlag==0)
	{
		return 0;
	}
	readFlag = spilu_checkDoubleInRange (fname, ivar, *permtol, 0, DBL_MAX);
	if (readFlag==0)
	{
		return 0;
	}
	return 1;
}

// 
// spilu_getArgumentBloc --
//   Get the bloc parameter.
//
// Arguments
//   fname (input) : the name of the Scilab function generating this error
//   ivar (input) : the index of the input variable
//   rhs (input): the number of input arguments
//   A (input) : the sparse matrix
//   bloc (output) : the value
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
// Check that drop is positive.
// If drop is not provided or is equal to [], then 
// set the default value, bloc=n.
//
int spilu_getArgumentBloc ( char * fname , int ivar , int rhs, SciSparse A , int * bloc )
{
	int defaultbloc;
	int readFlag;

	defaultbloc = A.m;

	readFlag = spilu_getScalarIntegerFromScalarDouble (fname, ivar, rhs, defaultbloc, bloc);
	if (readFlag==0)
	{
		return 0;
	}
	readFlag = spilu_checkIntegerInRange (fname, ivar, *bloc, 0, A.m);
	if (readFlag==0)
	{
		return 0;
	}
	return 1;
}

// 
// spilu_getArgumentLfil --
//   Get the lfil parameter.
//
// Arguments
//   fname (input) : the name of the Scilab function generating this error
//   ivar (input) : the index of the input variable
//   rhs (input): the number of input arguments
//   A (input) : the sparse matrix
//   lfil (output) : the value
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
// Check that drop is positive.
// If drop is not provided or is equal to [], then 
// set the default value, lfil = 1 + A.nel/A.n.
//
int spilu_getArgumentLfil ( char * fname , int ivar , int rhs, SciSparse A , int * lfil )
{
	int defaultlfil;
	int readFlag;

	defaultlfil = 1+((int) (A.nel/A.n));

	readFlag= spilu_getScalarIntegerFromScalarDouble (fname, ivar, Rhs, defaultlfil, lfil);
	if (readFlag==0)
	{
		return 0;
	}
	readFlag = spilu_checkIntegerInRange (fname, ivar, *lfil, 0, A.m);
	if (readFlag==0)
	{
		return 0;
	}
	return 1;
}

// spilu_allocLUArrays --
//   Allocate arrays of the LU decomposition 
//   necessary for the Fortran library.
//
// Arguments
// Input:
// fname : the name of the function
// nrows : the number of rows in the sparse matrix A
// nnzLU : an integer, the estimated number of nonzeros in the LU decomposition
//
// Output:
// ju: a n-by-1 matrix of doubles, pointer to the diagonal elements in alu, jlu
// alu: a (nnzLU+1)-by-1 matrix of doubles, the values of the ALU factors
// jlu: a (nnzLU+1)-by-1 matrix of doubles, pointers to the alu array
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
int spilu_allocLUArrays(char * fname, int nnzLU, int nrows, double ** alu, int ** jlu, int ** ju)
{
	*alu= (double *) malloc((nnzLU+1)*sizeof(double));
	if (*alu==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	*jlu= (int *) malloc((nnzLU+1)*sizeof(int));
	if (*jlu==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	*ju= (int *) malloc(nrows*sizeof(int));
	if (*ju==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	return 1;
}
// spilu_allocRowptr --
//   Allocate and compute the array ia.
//
// Arguments
// Input:
// fname : the name of the function
// A : a n-by-n sparse matrix
//
// Output:
// ia[0:nrows] (output) : an array with A.m+1 entries, pointers to the rows of A.
//            For i=1, 2,..., A.m, the value ia(i) is the first 
//            index of a for the row A(i,:).
//            In other words, for i=1, 2,..., A.m, 
//            a(ia(i):ia(i+1)-1) contains the nonzeros of the i-th row.
//            The number of nonzeros is : ia(A.m+1)-1
//            This is the "row_ptr" array in the CSR format.
//
// Description
// If there is a problem, generate an error and return 0.
// If there is no problem, returns 1.
//
int spilu_allocRowptr(char * fname, SciSparse A, int ** ia)
{
	int nrows = 0;
	nrows = A.m;

	*ia=(int *) malloc((nrows+1)*sizeof(int));
	if (*ia==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	spiluc_Sci2spk(A,*ia);
	return 1;
}